import numpy as np
from numpy.polynomial.polynomial import polyval
from scipy.special import kv


def Re(omega: float, w: float, nu: float) -> float:
    return omega * w**2 / (4 * nu)


def hydrodynamic_rectangular(Re: float) -> complex:
    return hydrodynamic_circular(Re) * correction_rectangular(Re)


def hydrodynamic_circular(Re: float) -> complex:
    z = -1j * np.sqrt(1j * Re)
    return 1 + 4 * kv(1, z) / (z * kv(0, z))


def correction_rectangular(Re: float) -> complex:
    tau = np.log10(Re)
    numer_real = polyval(
        tau, [0.91324, -0.48274, 0.46842, -0.12886, 0.044055, -0.0035117, 0.00069085]
    )
    denom_real = polyval(
        tau, [1, -0.56964, 0.48690, -0.13444, 0.045155, -0.0035862, 0.00069085]
    )
    numer_imag = polyval(
        tau, [-0.024134, -0.029256, 0.016294, -0.00010961, 0.000064577, -0.000044510]
    )
    denom_imag = polyval(
        tau, [1, -0.59702, 0.55182, -0.18357, 0.079156, -0.014369, 0.0028361]
    )
    return numer_real / denom_real + 1j * numer_imag / denom_imag


def resonance_curve(
    f: np.ndarray, f0: float, Q: float, P_dc: float, P_white: float
) -> np.ndarray:
    denominator = (f**2 - f0**2) ** 2 + f**2 * f0**2 / Q**2
    return P_white + P_dc * f0**4 / denominator



from pathlib import Path
import re

import numpy as np
import xarray as xr


def read_scope_file(filepath: Path | str) -> xr.DataArray:
    comments, data = read_csv_with_comments(filepath, comments="%", delimiter=";")
    label_list = comments[-1]
    names, units = list(zip(*[match_name_and_unit(l) for l in label_list]))
    output = xr.DataArray(
        data=data[:, 1],
        coords=[(names[0], data[:, 0], {"unit": units[0]})],
        name=names[1],
        attrs={"unit": units[1]},
    )
    return comments[:-1], output


def read_csv_with_comments(
    filepath: Path | str, comments: str = "%", delimiter: str = ";"
):
    with open(filepath, "r") as f:
        metadata = []
        for line in f:
            if not line.startswith(comments):
                break
            metadata.append(line.lstrip(comments).rstrip("\n").split(delimiter))
        body = np.loadtxt(f, dtype="float", comments=comments, delimiter=delimiter)
    return metadata, body


def match_name_and_unit(label: str) -> tuple[str, str]:
    match_result = re.match(r"\s*([\w+\s*]+) \((\S+)\)", label)
    name, unit = match_result.groups()
    return name.lower(), unit



from typing import Iterable
from itertools import chain

import xarray as xr
import numpy as np


def dataset_from_dataarrays(
    dataarrays: Iterable[xr.DataArray], name: str | None = None
) -> xr.Dataset:
    ds_dict = {da.name: da for da in dataarrays}
    return xr.Dataset(data_vars=ds_dict, attrs={"name": name})


def is_empty_dataset(dataset: xr.Dataset) -> bool:
    return len(dataset) == 0


def is_zero_dataset(dataset: xr.Dataset) -> bool:
    data, coords = dataset.data_vars.values(), dataset.coords.values()
    is_zero = all(is_zero_array(v.to_numpy()) for v in chain(data, coords))
    return is_zero


def is_zero_array(array: np.ndarray) -> bool:
    return np.all(array == 0.0)



import matplotlib


def configure_matplotlib_defaults(fontsize: int = 12) -> None:
    matplotlib.rc("font", size=fontsize)
    matplotlib.rc("axes", grid=True)
    matplotlib.rc("axes.spines", top=False, right=False)
    matplotlib.rc("grid", linestyle="--", color="lightgray")
    matplotlib.rc("savefig", facecolor="white")



from typing import TypeVar, Iterable, Callable, Sequence
from collections import defaultdict

K, V = TypeVar("K"), TypeVar("V")


def bin_key_value_list(key_value_pairs: Iterable[tuple[K, V]]) -> dict[K, list[V]]:
    binned_dict = defaultdict(list)
    for k, v in key_value_pairs:
        binned_dict[k].append(v)
    return binned_dict


def filter_dict(
    data_dict: dict[K, V], rejection_criteria: Sequence[Callable[[V], bool]]
) -> dict[K, V]:
    def filter_func(data: V) -> bool:
        return any(c(data) for c in rejection_criteria)

    return {k: v for k, v in data_dict.items() if not filter_func(v)}



from __future__ import annotations
from typing import Literal, BinaryIO, ByteString

import pydantic
from pydantic.dataclasses import dataclass
import numpy as np
import xarray as xr


@dataclass(frozen=True)
class ChannelHeader:
    version: int
    points: int
    lines: int
    frame: pydantic.constr(to_lower=True)
    curline: int

    dim0name: pydantic.constr(to_lower=True) = "dim0"
    dim0unit: str | None = None
    dim0range: float = 0.0
    dim0min: float = 0.0

    dim1name: pydantic.constr(to_lower=True) = "dim1"
    dim1unit: str | None = None
    dim1range: float = 0.0
    dim1min: float = 0.0

    dim2name: pydantic.constr(to_lower=True) = "dim2"
    dim2unit: str | None = None
    dim2range: float = 0.0
    dim2min: float = 0.0

    savemode: Literal["Binary"] = "Binary"
    savebits: pydantic.conint(multiple_of=8) = 16
    savesign: Literal["Signed", "Unsigned"] = "Signed"
    saveorder: Literal["Intel"] = "Intel"

    @property
    def measurement_type(self) -> str:
        return self.frame

    @property
    def data_name(self) -> str:
        return self.dim2name

    @property
    def data_unit(self) -> str:
        return self.dim2unit

    @property
    def data_shape(self) -> tuple[int, int]:
        return (self.points, self.lines)

    @property
    def data_dtype(self) -> np.dtype:
        endian = "<" if self.saveorder == "Intel" else ">"
        data_format = "i" if self.savesign == "Signed" else "u"
        num_bytes = self.savebits // 8
        return np.dtype(f"{endian}{data_format}{num_bytes}")

    @property
    def data_num_bytes(self) -> int:
        dtype = self.data_dtype
        return self.points * self.lines * dtype.itemsize

    @property
    def data_min(self) -> float:
        return self.dim2min

    @property
    def data_range(self) -> float:
        return self.dim2range


class Channel:
    def __init__(self, header: ChannelHeader, data: np.ndarray | None = None):
        self.header = header
        self.data = data

    def load_data_from_stream(self, binary_stream: BinaryIO) -> Channel:
        binary_data = binary_stream.read(self.header.data_num_bytes)
        raw_data = self._decode_binary_data(binary_data)
        data = self._scale_data(raw_data)
        return Channel(self.header, data)

    def _decode_binary_data(self, binary_data: ByteString) -> np.ndarray:
        return np.frombuffer(binary_data, dtype=self.header.data_dtype).reshape(
            self.header.data_shape
        )

    def _scale_data(self, raw_data: np.ndarray) -> np.ndarray:
        dtype_info = np.iinfo(self.header.data_dtype)
        data_max, data_min = dtype_info.max, dtype_info.min
        normalized_data = (np.float32(raw_data) - data_min) / (data_max - data_min)
        return normalized_data * self.header.data_range + self.header.data_min

    def to_dataarray(self) -> xr.DataArray:
        coords = self._make_dataarray_coords()
        dataarray = xr.DataArray(
            self.data,
            coords=list(coords),
            name=self.header.data_name,
            attrs={"unit": self.header.data_unit},
        )
        return dataarray

    def _make_dataarray_coords(self) -> tuple[xr.Variable, xr.Variable]:
        header = self.header
        data0 = np.linspace(0, header.dim0range, header.points) + header.dim0min
        data1 = np.linspace(0, header.dim1range, header.lines) + header.dim1min
        coord0 = xr.Variable(
            dims=header.dim0name, data=data0, attrs={"unit": header.dim0unit}
        )
        coord1 = xr.Variable(
            dims=header.dim1name, data=data1, attrs={"unit": header.dim1unit}
        )
        return coord0, coord1



import numpy as np
from skimage.morphology import opening


def spectral_subtraction(
    signal_spectrum: np.ndarray,
    reference_spectrum: np.ndarray,
    n_split: int = 1000,
    alpha0: float = 4,
    s: float = 20 / 3,
    beta: float = 2e-3,
) -> np.ndarray:
    signal_bins, reference_bins = np.array_split(
        signal_spectrum, n_split
    ), np.array_split(reference_spectrum, n_split)
    SNRs = [calculate_SNR(p, n) for p, n in zip(signal_bins, reference_bins)]
    alphas = _oversubtraction_factor(SNRs, alpha0, s)
    alpha_array = np.repeat(alphas, [len(p) for p in reference_bins])
    subtracted = signal_spectrum - alpha_array * reference_spectrum
    clipped = np.clip(subtracted, beta * signal_spectrum, np.inf)
    return clipped


def calculate_SNR(
    signal_power_spectrum: np.ndarray, noise_power_spectrum: np.ndarray
) -> np.ndarray:
    ratio = np.sum(signal_power_spectrum, axis=-1) / np.sum(
        noise_power_spectrum, axis=-1
    )
    return 10 * np.log10(ratio)


def _oversubtraction_factor(SNR: np.ndarray | float, alpha0=4, s=20 / 3) -> np.ndarray:
    SNR = np.array(SNR)
    alpha = alpha0 - np.clip(SNR, -5, 20) / s
    return alpha


def baseline_morphological(
    power_spectrum: np.ndarray, tol: float = 1e-9, max_iter: int = 200
) -> np.ndarray:
    baseline = np.array(power_spectrum)
    for i in range(max_iter):
        opened = opening(baseline, footprint=np.ones(i + 2))
        error = np.sum(baseline) - np.sum(opened)
        if error <= tol:
            break
        baseline = opened
    else:
        print(
            "Max iteration reached, may need to try again with a larger max_iter value or larger tol value"
        )
    return baseline


def baseline_rayleigh(power_spectrum, tol=0.999995, max_iter=10000) -> np.ndarray:

    power = np.array(power_spectrum)
    b = np.sum(power) / (2 * len(power))
    for _ in range(max_iter):
        threshold = 2 * b * (-np.log10(tol))
        power[power > threshold] = 0
        b_old, b = b, np.sum(power) / (2 * len(power))
        if b - b_old <= 1e-2:
            break
    else:
        print(
            "Max iteration reached, may need to try again with a larger max_iter value"
        )
    return threshold



from pathlib import Path
from typing import ByteString, BinaryIO, Mapping
import mmap
from io import BytesIO
from configparser import ConfigParser
import re

import xarray as xr

from jhelabtoolkit.io.nanosurf.nid_components import ChannelHeader, Channel
from jhelabtoolkit.utils.xarray import (
    dataset_from_dataarrays,
    is_empty_dataset,
    is_zero_dataset,
)
from jhelabtoolkit.utils.misc import bin_key_value_list, filter_dict


def read_nid(filepath: str | Path) -> tuple[ConfigParser, dict[str, xr.Dataset]]:
    with open(filepath, "rb") as file:
        with mmap.mmap(file.fileno(), length=0, access=mmap.ACCESS_READ) as memorymap:
            config_binary, data_binary = split_config_and_data(memorymap)
            config = parse_config(config_binary)
            channel_headers = get_channel_headers(config)
            data = parse_data(data_binary, channel_headers)
    data = filter_dict(data, rejection_criteria=(is_empty_dataset, is_zero_dataset))
    return config, data


def split_config_and_data(
    binary_fileio: BinaryIO, data_delimiter: ByteString = b"#!"
) -> tuple[ByteString, ByteString]:
    config_end = binary_fileio.find(data_delimiter)
    config = binary_fileio[:config_end]
    data = binary_fileio[config_end + len(data_delimiter) :]
    return config, data


def parse_config(config_binary: ByteString) -> ConfigParser:
    config_list = config_binary.decode().split("\r\n")
    config_parser = ConfigParser(allow_no_value=True)
    config_parser.read_file(config_list)
    return config_parser


def get_channel_headers(config: ConfigParser) -> list[ChannelHeader]:
    header_regex = re.compile("DataSet-\d+:\d+")
    headers = [
        ChannelHeader(**section)
        for name, section in config.items()
        if re.match(header_regex, name) is not None
    ]
    return headers


def parse_data(
    data_binary: ByteString, channel_headers: list[ChannelHeader]
) -> dict[str, xr.Dataset]:
    data_stream = BytesIO(data_binary)
    parsed_channels = [parse_channel(data_stream, hdr) for hdr in channel_headers]
    channel_group_dict = bin_key_value_list(parsed_channels)
    data = {
        group_name: dataset_from_dataarrays(group_data)
        for group_name, group_data in channel_group_dict.items()
    }
    return data


def parse_channel(
    binary_stream: BinaryIO, header: ChannelHeader
) -> tuple[str, xr.DataArray]:
    meas_type = header.measurement_type
    channel_empty = Channel(header)
    channel = channel_empty.load_data_from_stream(binary_stream)
    return meas_type, channel.to_dataarray()


def filter_dataset_info(dataset_info: Mapping[str, str]):
    dummy_key_regex = re.compile("-- \s+ --")
    return {
        k: v
        for k, v in dataset_info.items()
        if re.match(dummy_key_regex, k) is not None
    }
